const {
  time,
  loadFixture,
} = require('@nomicfoundation/hardhat-toolbox/network-helpers');
const { expect } = require('chai');
const { ethers } = require('hardhat');

describe('Forum', () => {
  async function deploy() {
    const [account1, account2, account3] = await ethers.getSigners();
    const DAO = await ethers.getContractFactory('DAO');
    const dao = await DAO.deploy();
    return {
      dao, account1, account2, account3,
    };
  }
  let dao;
  let account1;
  let account2;
  let account3;
  const POOL_DURATION = 3600; // 1 hour
  const POOL_FEE = 100;
  const emptyCallbackData = ethers.AbiCoder.defaultAbiCoder().encode([], []);

  const initiateValidationPool = ({
    postId, duration,
    quorum, winRatio, bindingPercent,
    redistributeLosingStakes, callbackOnValidate,
    callbackData, fee,
  } = {}) => dao.initiateValidationPool(
    postId,
    duration ?? POOL_DURATION,
    quorum ?? [1, 10],
    winRatio ?? [1, 2],
    bindingPercent ?? 100,
    redistributeLosingStakes ?? true,
    callbackOnValidate ?? false,
    callbackData ?? emptyCallbackData,
    { value: fee ?? POOL_FEE },
  );

  describe('Post', () => {
    beforeEach(async () => {
      ({
        dao, account1, account2, account3,
      } = await loadFixture(deploy));
    });

    it('should be able to add a post', async () => {
      const contentId = 'some-id';
      await expect(dao.addPost(account1, contentId, [])).to.emit(dao, 'PostAdded').withArgs('some-id');
      const post = await dao.posts(contentId);
      expect(post.author).to.equal(account1);
      expect(post.sender).to.equal(account1);
      expect(post.id).to.equal(contentId);
    });

    it('should be able to add a post on behalf of another account', async () => {
      const contentId = 'some-id';
      await dao.addPost(account2, contentId, []);
      const post = await dao.posts(contentId);
      expect(post.author).to.equal(account2);
      expect(post.sender).to.equal(account1);
      expect(post.id).to.equal(contentId);
    });

    it('should be able to donate reputation via citations', async () => {
      await dao.addPost(account1, 'content-id', []);
      await dao.addPost(account2, 'second-content-id', [{ weightPercent: 50, targetPostId: 'content-id' }]);
      await initiateValidationPool({ postId: 'second-content-id' });
      const pool = await dao.validationPools(0);
      expect(pool.postId).to.equal('second-content-id');
      await dao.evaluateOutcome(0);
      expect(await dao.balanceOf(account1)).to.equal(50);
      expect(await dao.balanceOf(account2)).to.equal(50);
    });

    it('should be able to leach reputation via citations', async () => {
      await dao.addPost(account1, 'content-id', []);
      expect((await dao.posts('content-id')).reputation).to.equal(0);
      await initiateValidationPool({ postId: 'content-id' });
      await dao.evaluateOutcome(0);
      expect(await dao.balanceOf(account1)).to.equal(100);
      expect((await dao.posts('content-id')).reputation).to.equal(100);
      await dao.addPost(account2, 'second-content-id', [{ weightPercent: -50, targetPostId: 'content-id' }]);
      expect((await dao.posts('second-content-id')).reputation).to.equal(0);
      await initiateValidationPool({ postId: 'second-content-id' });
      const pool = await dao.validationPools(1);
      expect(pool.postId).to.equal('second-content-id');
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(1);
      expect(await dao.balanceOf(account1)).to.equal(50);
      expect(await dao.balanceOf(account2)).to.equal(150);
      expect((await dao.posts('content-id')).reputation).to.equal(50);
      expect((await dao.posts('second-content-id')).reputation).to.equal(150);
    });

    it('should be able to redistribute power via citations', async () => {
      await dao.addPost(account1, 'content-id', []);
      await initiateValidationPool({ postId: 'content-id' });
      await dao.evaluateOutcome(0);
      expect(await dao.balanceOf(account1)).to.equal(100);
      await dao.addPost(account2, 'second-content-id', []);
      expect(await dao.balanceOf(account2)).to.equal(0);
      await dao.addPost(account3, 'third-content-id', [
        { weightPercent: -100, targetPostId: 'content-id' },
        { weightPercent: 100, targetPostId: 'second-content-id' },
      ]);
      await initiateValidationPool({ postId: 'third-content-id' });
      const pool = await dao.validationPools(1);
      expect(pool.postId).to.equal('third-content-id');
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(1);
      expect(await dao.balanceOf(account1)).to.equal(0);
      expect(await dao.balanceOf(account2)).to.equal(200);
      expect(await dao.balanceOf(account3)).to.equal(0);
    });

    it('should be able to reverse a negative citation with a negative citation', async () => {
      await dao.addPost(account1, 'content-id', []);
      await initiateValidationPool({ postId: 'content-id' });
      await dao.evaluateOutcome(0);
      expect(await dao.balanceOf(account1)).to.equal(100);
      await dao.addPost(account2, 'second-content-id', [{ weightPercent: -100, targetPostId: 'content-id' }]);
      await initiateValidationPool({ postId: 'second-content-id' });
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(1);
      expect(await dao.balanceOf(account1)).to.equal(0);
      expect(await dao.balanceOf(account2)).to.equal(200);
      await dao.addPost(account3, 'third-content-id', [{ weightPercent: -100, targetPostId: 'second-content-id' }]);
      await initiateValidationPool({ postId: 'third-content-id' });
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(2);
      expect(await dao.balanceOf(account1)).to.equal(100);
      expect(await dao.balanceOf(account2)).to.equal(0);
      expect(await dao.balanceOf(account3)).to.equal(200);
    });

    it('forum reputation rewards are shared with validation pool policing rewards', async () => {
      await dao.addPost(account1, 'content-id', []);
      await initiateValidationPool({ postId: 'content-id' });
      await dao.evaluateOutcome(0);
      expect(await dao.balanceOf(account1)).to.equal(100);
      await dao.addPost(account2, 'second-content-id', []);
      await initiateValidationPool({ postId: 'second-content-id' });
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(1);
      expect(await dao.balanceOf(account1)).to.equal(100);
      expect(await dao.balanceOf(account2)).to.equal(100);
      await dao.addPost(account3, 'third-content-id', [{ weightPercent: -100, targetPostId: 'second-content-id' }]);
      await initiateValidationPool({ postId: 'third-content-id' });
      await dao.stakeOnValidationPool(2, 100, true);
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(2);
      expect(await dao.balanceOf(account1)).to.equal(133);
      expect(await dao.balanceOf(account2)).to.equal(33);
      expect(await dao.balanceOf(account3)).to.equal(134);
    });

    it('should limit effects of negative references on prior positive references', async () => {
      console.log('First post');
      await dao.addPost(account1, 'content-id', []);
      await initiateValidationPool({ postId: 'content-id' });
      await dao.evaluateOutcome(0);
      expect(await dao.balanceOf(account1)).to.equal(100);
      await dao.addPost(account2, 'second-content-id', [{ weightPercent: 50, targetPostId: 'content-id' }]);
      await initiateValidationPool({ postId: 'second-content-id' });
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(1);
      expect(await dao.balanceOf(account1)).to.equal(150);
      expect(await dao.balanceOf(account2)).to.equal(50);
      await dao.addPost(account3, 'third-content-id', [{ weightPercent: -100, targetPostId: 'second-content-id' }]);
      await initiateValidationPool({ postId: 'third-content-id', fee: 200 });
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(2);
      expect(await dao.balanceOf(account1)).to.equal(100);
      expect(await dao.balanceOf(account2)).to.equal(0);
      expect(await dao.balanceOf(account3)).to.equal(300);
    });

    it('should limit effects of negative references on prior negative references', async () => {
      await dao.addPost(account1, 'content-id', []);
      await initiateValidationPool({ postId: 'content-id' });
      await dao.evaluateOutcome(0);
      expect(await dao.balanceOf(account1)).to.equal(100);
      await dao.addPost(account2, 'second-content-id', [{ weightPercent: -100, targetPostId: 'content-id' }]);
      await initiateValidationPool({ postId: 'second-content-id' });
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(1);
      expect(await dao.balanceOf(account1)).to.equal(0);
      expect(await dao.balanceOf(account2)).to.equal(200);
      await dao.addPost(account3, 'third-content-id', [{ weightPercent: -100, targetPostId: 'second-content-id' }]);
      await initiateValidationPool({ postId: 'third-content-id', fee: 200 });
      await time.increase(POOL_DURATION + 1);
      await dao.evaluateOutcome(2);
      expect(await dao.balanceOf(account1)).to.equal(100);
      expect(await dao.balanceOf(account2)).to.equal(0);
      expect(await dao.balanceOf(account3)).to.equal(300);
    });

    it('should enforce depth limit', async () => {
      await dao.addPost(account1, 'content-id-1', []);
      await dao.addPost(account1, 'content-id-2', [{ weightPercent: 100, targetPostId: 'content-id-1' }]);
      await dao.addPost(account1, 'content-id-3', [{ weightPercent: 100, targetPostId: 'content-id-2' }]);
      await dao.addPost(account1, 'content-id-4', [{ weightPercent: 100, targetPostId: 'content-id-3' }]);
      await initiateValidationPool({ postId: 'content-id-4' });
      await dao.evaluateOutcome(0);
      const posts = await Promise.all([
        await dao.posts('content-id-1'),
        await dao.posts('content-id-2'),
        await dao.posts('content-id-3'),
        await dao.posts('content-id-4'),
      ]);
      expect(posts[0].reputation).to.equal(0);
      expect(posts[1].reputation).to.equal(100);
      expect(posts[2].reputation).to.equal(0);
      expect(posts[3].reputation).to.equal(0);
    });

    it('should be able to incinerate reputation', async () => {
      await dao.addPost(account1, 'content-id-1', [
        {
          weightPercent: 50,
          targetPostId: '',
        },
      ]);
      await initiateValidationPool({ postId: 'content-id-1' });
      expect(await dao.totalSupply()).to.equal(100);
      await dao.evaluateOutcome(0);
      expect((await dao.posts('content-id-1')).reputation).to.equal(50);
      expect(await dao.totalSupply()).to.equal(50);
    });

    describe('negative citation of a post, the author having already staked and lost reputation', async () => {
      beforeEach(async () => {
        await dao.addPost(account1, 'content-id', []);
        await initiateValidationPool({ postId: 'content-id' });
        await dao.evaluateOutcome(0);
        expect(await dao.balanceOf(account1)).to.equal(100);
        expect(await dao.totalSupply()).to.equal(100);
        expect((await dao.posts('content-id')).reputation).to.equal(100);

        await dao.addPost(account2, 'second-content-id', []);
        await initiateValidationPool({ postId: 'second-content-id' });
        await time.increase(POOL_DURATION + 1);
        await dao.evaluateOutcome(1);
        expect(await dao.balanceOf(account1)).to.equal(100);
        expect(await dao.balanceOf(account2)).to.equal(100);
        expect(await dao.totalSupply()).to.equal(200);
        expect((await dao.posts('content-id')).reputation).to.equal(100);
        expect((await dao.posts('second-content-id')).reputation).to.equal(100);

        // account1 stakes and loses
        await initiateValidationPool({ postId: 'second-content-id' });
        await dao.stakeOnValidationPool(2, 50, true);
        await dao.connect(account2).stakeOnValidationPool(2, 60, false);
        await time.increase(POOL_DURATION + 1);
        await dao.evaluateOutcome(2);
        expect(await dao.balanceOf(account1)).to.equal(50);
        expect(await dao.balanceOf(account2)).to.equal(250);
        expect(await dao.totalSupply()).to.equal(300);
        expect((await dao.posts('content-id')).reputation).to.equal(100);
        expect((await dao.posts('second-content-id')).reputation).to.equal(100);
      });

      it('author and post rep can be completely destroyed', async () => {
        // account1's post is later strongly negatively referenced
        await dao.addPost(account3, 'third-content-id', [{ weightPercent: -100, targetPostId: 'content-id' }]);
        await initiateValidationPool({ postId: 'third-content-id', fee: 200 });
        await time.increase(POOL_DURATION + 1);
        await dao.evaluateOutcome(3);
        expect(await dao.balanceOf(account1)).to.equal(0);
        expect(await dao.balanceOf(account2)).to.equal(250);
        expect(await dao.balanceOf(account3)).to.equal(250);
        expect(await dao.totalSupply()).to.equal(500);
        expect((await dao.posts('content-id')).reputation).to.equal(0);
        expect((await dao.posts('second-content-id')).reputation).to.equal(100);
        expect((await dao.posts('third-content-id')).reputation).to.equal(250);
      });

      it('author rep can be destroyed while some post rep remains', async () => {
        // account1's post is later strongly negatively referenced
        await dao.addPost(account3, 'third-content-id', [{ weightPercent: -100, targetPostId: 'content-id' }]);
        await initiateValidationPool({ postId: 'third-content-id', fee: 70 });
        await time.increase(POOL_DURATION + 1);
        await dao.evaluateOutcome(3);
        expect(await dao.totalSupply()).to.equal(370);
        expect(await dao.balanceOf(account1)).to.equal(0);
        expect(await dao.balanceOf(account2)).to.equal(250);
        expect(await dao.balanceOf(account3)).to.equal(120);
        expect((await dao.posts('content-id')).reputation).to.equal(30);
        expect((await dao.posts('second-content-id')).reputation).to.equal(100);
        expect((await dao.posts('third-content-id')).reputation).to.equal(120);
      });

      it('author rep can be destroyed while some post rep remains (odd amount)', async () => {
        // account1's post is later strongly negatively referenced
        await dao.addPost(account3, 'third-content-id', [{ weightPercent: -100, targetPostId: 'content-id' }]);
        await initiateValidationPool({ postId: 'third-content-id', fee: 75 });
        await time.increase(POOL_DURATION + 1);
        await dao.evaluateOutcome(3);
        expect(await dao.totalSupply()).to.equal(375);
        expect(await dao.balanceOf(account1)).to.equal(0);
        expect(await dao.balanceOf(account2)).to.equal(250);
        expect(await dao.balanceOf(account3)).to.equal(125);
        expect((await dao.posts('content-id')).reputation).to.equal(25);
        expect((await dao.posts('second-content-id')).reputation).to.equal(100);
        expect((await dao.posts('third-content-id')).reputation).to.equal(125);
      });
    });
  });
});
